@@ -18,26 +18,11 @@ import sys
 
 def pidof(pgname):
     pids = []
-    for proc in psutil.process_iter():
-        with proc.oneshot():
-            # search for matches in the process name and cmdline
-            try:
-                name = proc.name()
-            except psutil.Error:
-                pass
-            else:
-                if name == pgname:
-                    pids.append(str(proc.pid))
-                    continue
-
-            try:
-                cmdline = proc.cmdline()
-            except psutil.Error:
-                pass
-            else:
-                if cmdline and cmdline[0] == pgname:
-                    pids.append(str(proc.pid))
-
+    for proc in psutil.process_iter(attrs=['name', 'cmdline']):
+        # search for matches in the process name and cmdline
+        if proc.info['name'] == pgname or \
+                proc.info['cmdline'] and proc.info['cmdline'][0] == pgname:
+            pids.append(str(proc.pid))
     return pids
 
 
